var SQLITE3 = require("sqlite3").verbose();
var FS = require("fs");
var TICKER_FREQUENCY = 20 * 1000;
var MEMORY_CLEAR_FREQUENCY = 120 * 1000;

function checkSchema(db, oncomplete)
{
	var st = db.prepare('SELECT sql FROM sqlite_master WHERE type = "table"  and name = "store"');
	st.get(function(error, row)
	{
		if (error)
			throw("invalid db schema");

		if (typeof row["sql"] == "string")
		{
			if (row["sql"].indexOf('PRIMARY KEY(`key`)') == -1)
			{
				throw("invalid db schema");
			}
			else
			{
				oncomplete();
				return;
			}
		}
		throw("invalid db schema");
	});
}

function createDB(db, oncomplete)
{
	var sql = 'CREATE TABLE `store` ( `key` TEXT, `value` TEXT, PRIMARY KEY(`key`) )';
	db.run(sql, oncomplete);
}

exports.kv = function(path, onload)
{
	var self = this;
	if (typeof onload == 'undefined')
		onload = function(){};

	self.path = path;
	self.is_path_exists = FS.existsSync(path);
	self.dataMap = new Map();
	self.notSyncedKeys = new Set();
	self.lastOperationTimeMap = new Map();
	self.stmt = {};
	self.db = new SQLITE3.Database(path, SQLITE3.OPEN_READWRITE | SQLITE3.OPEN_CREATE, function(err)
	{
		if (err)
			throw("unable to open db " + path);

		if (self.is_path_exists)
		{
			checkSchema(self.db, function()
			{
				self.stmt.insert = self.db.prepare("REPLACE INTO store (key,value) VALUES (?,?)");
				self.stmt.select = self.db.prepare("SELECT value FROM store WHERE key = ?");
				onload();
			});
		}
		else
		{
			createDB(self.db, function()
			{
				self.stmt.insert = self.db.prepare("REPLACE INTO store (key,value) VALUES (?,?)");
				self.stmt.select = self.db.prepare("SELECT value FROM store WHERE key = ?");
				onload();
			})
		}
	});

	
	self.set = function(key, value)
	{
		self.notSyncedKeys.add(key);
		self.lastOperationTimeMap.set(key, Date.now());
		self.dataMap.set(key, value);
	}

	self.get = function(key, oncomplete)
	{
		if (typeof oncomplete == 'undefined')
			oncomplete = function(){}
		
		var v;
		if (self.dataMap.has(key))
		{
			v = self.dataMap.get(key);
			self.lastOperationTimeMap.set(key, Date.now());
			setImmediate(function(){oncomplete(v);});
		}
		else
		{
			self.stmt.select.get(key , function(error,row)
			{
				if (typeof row == "undefined")
				{
					v = undefined;
				}
				else
				{
					v = row.value;
					self.dataMap.set(key,v);
					self.lastOperationTimeMap.set(key, Date.now());
				}
				setImmediate(function(){oncomplete(v);});
			});
		}
	}

	self.flush = function()
	{
		self.db.serialize(function()
		{
			self.db.run("BEGIN TRANSACTION");
			var counter = 0;
			self.notSyncedKeys.forEach(function(key)
			{
				var value = self.dataMap.get(key);
				self.stmt.insert.run(key, value);
				counter++;
			});
			self.db.run("COMMIT");
			if (counter > 0)
				console.log(counter + " write query executed to " + self.path);

			self.notSyncedKeys.clear();
		});
		clear();
	}

	setInterval(self.flush, TICKER_FREQUENCY);
	function clear()
	{
		var now = Date.now();
		self.lastOperationTimeMap.forEach(function(time, key)
		{
			var delta = now - time;
			if (delta > MEMORY_CLEAR_FREQUENCY)
			{
				del(key);
			}
		});
	}

	function del(key)
	{
		self.dataMap.delete(key);
		self.lastOperationTimeMap.delete(key);
	}

	return self;
}